/**
 * Copyright (C) 2012 Matteo Tamigi
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, 
 * USA.
*/

#include "PDFSec.hpp"

/** sets the number of bytes to decrypt for partial test.
    Three should be a good number for this as this mean a match should only
    happen every 256^3=16777216 check and that should be unique enough to 
    motivate a full retry on that entry.
 */
#define PARTIAL_TEST_SIZE 3

static void
printHexString(vector<uint8_t> str) {
  for(int i=0;i<str.size();i++)
    printf("%x ",str[i]);
  printf("\n");
}

static void
printHexString(const uint8_t * str, int size) {
  for(int i=0;i<size;i++)
    printf("%x ",str[i]);
  printf("\n");
}

//Costanti di cifratura per inizializzare i vari algoritmi di cifratura
const uint8_t pad[] = {
  0x28,0xBF,0x4E,0x5E,0x4E,0x75,0x8A,0x41,
  0x64,0x00,0x4E,0x56,0xFF,0xFA,0x01,0x08,
  0x2E,0x2E,0x00,0xB6,0xD0,0x68,0x3E,0x80,
  0x2F,0x0C,0xA9,0xFE,0x64,0x53,0x69,0x7A
};

//Funzioni per la cifratura/decifratura delle varie versioni della user_password
bool PDFSec :: isUserPasswordV1R2 () {
    md5(key,sizeKey,digest);                                  //Cifro la key con l'md5
    //Eseguo un test sui primi 4 caratteri 
    Rc4_40( u_string , PARTIAL_TEST_SIZE , digest , tmp );    //Cifro la u_str con le 3 cifre prese prima
    //Devo ottenere in dec il pad
    for(a = 0; a < PARTIAL_TEST_SIZE; a++)
        if(tmp[a] != pad[a])
            return false;
    //------------------------------------------------------------------------//
    //Test superato, proseguo con quello completo (prendendo solo i primi 5 caratteri dell'md5)
    Rc4_40( u_string , 32 , digest , tmp);    //Cifro la u_str con le 5 cifre prese prima
    //Devo ottenere in dec il pad
    for(a = 0; a < 32; a++)
        if(tmp[a] != pad[a])
            return false;
    //Superato il test, la chiave  quella giusta
    return true;
}

bool PDFSec :: isUserPasswordV2R3 () {
    //Cifro la key per ottenere cifr
    md5(key,sizeKey,digest);                //Cifro la key con l'md5
    md5_50(digest);                           //Cifro in md5 altre 50 volte (nel dubbio :D)
    //Eseguo un test sulle prime 3 cifre (per evitare di farlo tutto), se lo passa
    //eseguo il test completo
    for(a = 0; a < PARTIAL_TEST_SIZE; a++)
        tmp[a] = u_string[a];
    //Cifro con l'RC4 su 3 cifre
    for(a = 19; a >= 0; --a) {
        for(b = 0; b < length; ++b)
            tmpkey[b] = digest[b] ^ a;
        Rc4_128(tmp,PARTIAL_TEST_SIZE,tmpkey,tmp);
    }
    //Confronto il risultato atteso cifr con la user_pswd cifrata
    for(a = 0; a < PARTIAL_TEST_SIZE; a++)
        if(tmp[a] != cifr[a])
            return false;
    //------------------------------------------------------------------------//
    //Il test precedente ha avuto buon esito, ora effettuo la verifica totale
    for(a = 0; a < 16; a++)
        tmp[a] = u_string[a];
    //Cifro con l'RC4 su tutte e 16 le cifre
    for(a = 19; a >= 0; --a) {
        for(b = 0; b < length; ++b)
            tmpkey[b] = digest[b] ^ a;
        Rc4_128(tmp,16,tmpkey,tmp);
    }
    //Confronto il risultato atteso cifr con la user_pswd cifrata
    for(a = 0; a < 16; a++)
        if(tmp[a] != cifr[a])
            return false;
    //Test finale passato!
    return true;
}

//Funzioni per la cifratura/decifratura delle varie versioni della owner_password
//Algoritmo 3.4
bool PDFSec :: isOwnerPasswordV1R2 () {
    //Cifro la password corrente con l'md5
    md5(key,32,digest);
    //Uso il risultato sopra come chiave per cifrare (RC4 a 40 bit) la o_string 
    Rc4_40(o_string,32,digest,key2);
    //Effettuo un md5 sulla key2 (che contiene l'rc4 dell'md5 della password)
    md5(key2,sizeKey,digest);    
    //Vabb... andiamo avanti cos insomma!
    Rc4_40(u_string,32,digest,tmp);
    //Verifico che ci che si trova nel tmp corrisponda al pad
    for(a = 0; a < 32; a++)
        if( tmp[a] != pad[a] )
            return false;
    //Il controllo ha avuto successo! La password cercata  corretta!
    return true;
}

//Reverse dell'algoritmo 3.4 (per cifrare la o_string) nel caso la user_password
//sia nulla
bool PDFSec :: isOwnerPasswordV1R2_nuser () {
    md5(key,32,digest);
    Rc4_40(o_string,32,digest,key2);
    //Nella key2 dovrei ottenere il pad
    for(a = 0; a < 32; a++)
        if( key2[a] != pad[a] )
            return false;
    return true;
}

bool PDFSec :: isOwnerPasswordV2R3 () {
    //Effettuo una copia della key (mi serve :\)
    for(a = 0; a < 32; a++)
        key2[a] = key[a];
    //Cifro la padded-password con l'md5
    md5(key,32,digest);
    //Do altre 50 passate per sicurezza
    md5_50(digest);
    //Copio la o_string su tmp per poterla modificare
    for(a = 0; a < 32; a++)
        tmp[a] = o_string[a];
    //Cifro con l'RC4 su tutte e 16 le cifre
    for(a = 19; a >= 0; --a) {
        for(b = 0; b < length; ++b)
            tmpkey[b] = digest[b] ^ a;
        Rc4_128(tmp,32,tmpkey,tmp);
    }
    //in tmp dovrei ottenere la padded-password della V2R3
    for(a = 0; a < 32; a++)
        key[a] = tmp[a];
    //Verifico sia quella corretta
    if(isUserPasswordV2R3())
        return true;
    //ripristino la copia della key
    for(a = 0; a < 32; a++)
        key[a] = key2[a];
    return false;
}

bool PDFSec :: isOwnerPasswordV2R3_nuser () {
    //Cifro la padded-password con l'md5
    md5(key,32,digest);
    //Do altre 50 passate per sicurezza
    md5_50(digest);
    //Copio la o_string su tmp per poterla modificare
    for(a = 0; a < 32; a++)
        tmp[a] = o_string[a];
    //Cifro con l'RC4 su 3 le cifre
    for(a = 19; a >= 0; --a) {
        for(b = 0; b < length; ++b)
            tmpkey[b] = digest[b] ^ a;
        Rc4_128(tmp,PARTIAL_TEST_SIZE,tmpkey,tmp);
    }
    for(a = 0; a < PARTIAL_TEST_SIZE; a++)
        if( tmp[a] != pad[a] )
            return false;
    //Tutto ok, faccio il controllo completo
    for(a = 19; a >= 0; --a) {
        for(b = 0; b < length; ++b)
            tmpkey[b] = digest[b] ^ a;
        Rc4_128(tmp,32,tmpkey,tmp);
    }
    for(a = 0; a < 32; a++)
        if( tmp[a] != pad[a] )
            return false;
    //Password corretta
    return true;
}


//Funzioni della classe

void PDFSec :: initCracking ( const string usr_pswd = "" ) {
    //Verifico che siano state trovate informazioni riguardo alla cifratura
    if( !infoPdf ) {
        cerr << "[initCracking] No file loaded!" << endl;
        return;
    } else if( ERRNOENCINFO == err ) {
        cerr << "[initCracking] No encryption infos found from selected file!" << endl;
        return;
    }
    //Pulisco la memoria
    clearCracking();
    //Stabilisco la versione della cifratura ed effettuo le operazioni preliminari:
    //-Verifico che i parametri in ingresso siano corretti
    //-Se esiste la user password, verifico che sia quella corretta
    //-Genero il dizionario per le password di bruteforcing
    //------------------------------------------------------------------------//
    
    /*--------------------------------------------------------------------------
    Standard di cifratura PDF:
    V:  0   Algoritmi non pubblicati e/o non pi supportati
        1   RC4 a 40-bit con parole da 32 caratteri Latin-1                                 R: 2
        2   RC4 a lunghezza variabile tra 40 e 128 bit con parole da 32 caratteri Latin-1   R: 3
        3   Algoritmo non documentato con dimensione variabile tra 40 e 128 bit             R: 3
        4   Algoritmo che utilizza specifiche implementazioni (stronze)
    --------------------------------------------------------------------------*/
    
    //Riempio la chiave da cifrare
    // password_padded + o_string + permissions + fileID [ (se la revision  > 3) + encryptMetaData]
    vector<uint8_t> keyy;
    //Se c' la user password la inserisco e la utilizzo
    if(usr_pswd != "") {
        userPswd = usr_pswd;
        isUserPswd = true;
        if( 32 < userPswd.length())
            userPswd.erase(userPswd.begin()+32,userPswd.end());    //Se la password  superiore ai 32 caratteri, la taglio
        //Riempio la chiave di cifratura con la user_pswd
        for(unsigned i = 0; i < userPswd.length(); i++)
            keyy.push_back((uint8_t)userPswd[i]);
    } else 
        isUserPswd = false;
    
    for(unsigned i = 0; i < 32-userPswd.length(); i++)
        keyy.push_back(pad[i]);
    for(unsigned i = 0; i < 32; i++)
        keyy.push_back(infoPdf->o_string[i]);
    keyy.push_back(infoPdf->permissions & 0xff);
    keyy.push_back((infoPdf->permissions >> 8) & 0xff);
    keyy.push_back((infoPdf->permissions >> 16) & 0xff);
    keyy.push_back((infoPdf->permissions >> 24) & 0xff);
    for(unsigned i = 0; i < infoPdf->fileID.size(); i++)
        keyy.push_back(infoPdf->fileID[i]);
    if(infoPdf->revision > 3 && !infoPdf->encryptMetaData) {
        keyy.push_back(0xff);
        keyy.push_back(0xff);
        keyy.push_back(0xff);
        keyy.push_back(0xff);
    }
    //Copio i risultati sul vettore key (che utilizzer per il cracking)
    key = new uint8_t[keyy.size()];
    sizeKey = keyy.size();
    for(unsigned i = 0; i < keyy.size(); i++) {
        key[i] = keyy[i];
    }
    //Genero la chiave da usare per la owner_password
    key2 = new uint8_t[keyy.size()];
    for(unsigned i = 0; i < 32; i++)
        key2[i] = pad[i];
    for(unsigned i = 32; i < keyy.size(); i++)
        key2[i] = keyy[i];
    
    u_string = new uint8_t[infoPdf->u_string.size()];
    o_string = new uint8_t[infoPdf->o_string.size()];
    for(unsigned i = 0; i < infoPdf->u_string.size(); i++)
        u_string[i] = infoPdf->u_string[i];
    for(unsigned i = 0; i < infoPdf->o_string.size(); i++)
        o_string[i] = infoPdf->o_string[i];
    //Genero il vettore per il risultato dell'md5
    digest = new uint8_t[16];
    //Verifico la revision per settare le varie variabili per il cracking
    switch( infoPdf->revision ) {
        case 2:
            tmp = new uint8_t[32];
            break;
        case 3:
            //Genero la chiave di test (diversa dal pad)
            //md5 ( pad + fileID )
            uint8_t * tempCifr;
            cifr = new uint8_t[32 + infoPdf->fileID.size()];
            for(unsigned i = 0; i < 32; i++)
                cifr[i] = pad[i];
            for(unsigned i = 0; i < infoPdf->fileID.size(); i++)
                cifr[32 + i] = infoPdf->fileID[i];
            md5(cifr,32 + infoPdf->fileID.size(),digest);
            delete [] cifr;
            cifr = new uint8_t[16];
            for(unsigned i = 0; i < 16; i++)
                cifr[i] = digest[i];
            //Genero le variabili di lavoro
            length = infoPdf->length/8;
            tmp = new uint8_t[32];
            tmpkey = new uint8_t[length];
            break;
    }
    //Come controllo finale, verifico che l'eventuale user_password fornita
    //sia corretta, altrimenti insulto l'utente e procedo con il cracking
    //della user_password
    if( isUserPswd && (( infoPdf->revision == 2 && !isUserPasswordV1R2() ) ||
                      ( infoPdf->revision == 3 && !isUserPasswordV2R3() )) ) {
        cerr << "WARNING: user_password specified isn't correctly! Re-initializing..." << endl;
        initCracking();
    }
}

//Funzione di testing, verr sostituita con un sistema di thread per l'elaborazione
//parallela su pi core (tanti thread quanti core fisici/virtuali) e/o tramite le
//varie tecnologie di GPGPU se possibile
string PDFSec :: crackUserPassword ( unsigned min_pw = 1, unsigned max_pw = 5,
                                     const string chst = STD_CHARSET , unsigned * st = NULL ) {
    //Verifico che sia stato inizializzato l'ambiente di cracking
    if( key == NULL ) {
        cerr << "WARNING: Cracking workspace isn't initialized!" << endl;
        return "";
    }
    //Verifico se la user_password sia quella nulla (per sicurezza)
    if( 1 == infoPdf->version && 2 == infoPdf->revision ) {
        if( !isUserPswd && isUserPasswordV1R2() ) {
            isUserPswd = true;
            userPswd = "";
            return "";
        }
        SetCharset(chst);
        SetState(min_pw,max_pw,st);    
        //Faccio partire il ciclo di cracking
        while ( !isUserPasswordV1R2() ) {
            SetNextPswd();
        }
        isUserPswd = true;
        userPswd = GetCurrentPswd();
    } else if ( 2 == infoPdf->version && 3 == infoPdf->revision ) {
        if( !isUserPswd && isUserPasswordV2R3() ) {
            isUserPswd = true;
            userPswd = "";
            return "";
        }
        SetCharset(chst);
        SetState(min_pw,max_pw,st);    
        //Faccio partire il ciclo di cracking
        while ( !isUserPasswordV2R3() ) {
            SetNextPswd();
        }
        isUserPswd = true;
        userPswd = GetCurrentPswd();
    }
    clearPassState ();
    return userPswd;
}

//Funzione di testing, verr sostituita con un sistema di thread per l'elaborazione
//parallela su pi core (tanti thread quanti core fisici/virtuali) e/o tramite le
//varie tecnologie di GPGPU se possibile
string PDFSec :: crackOwnerPassword ( unsigned min_pw = 1, unsigned max_pw = 5,
                                     const string chst = STD_CHARSET , unsigned * st = NULL ) {
    //Verifico che sia stato inizializzato l'ambiente di cracking
    if( key == NULL ) {
        cerr << "WARNING: Cracking workspace isn't initialized!" << endl;
        return "";
    }
    //Eseguo il ciclo di cracking della owner password
    if( 1 == infoPdf->version && 2 == infoPdf->revision ) {
        //Verifico se la user_password  nulla per poter usare il reverse
        if( !isUserPswd && isUserPasswordV1R2() ) {
                isUserPswd = true;
                userPswd = "";
        }
        SetCharset(chst);
        SetState(min_pw,max_pw,st);
        //Nel caso non vi sia una user_password posso usare un reverse pi veloce
        if( isUserPswd && userPswd == "" ) {
            //Faccio partire il ciclo di cracking
            while ( !isOwnerPasswordV1R2_nuser() ) {
                SetNextPswd();
            }
        } else {    //Altrimenti DEVO utilizzare il metodo standard
            //Faccio partire il ciclo di cracking
            while ( !isOwnerPasswordV1R2() )
                SetNextPswd();
        }
    } else if ( 2 == infoPdf->version && 3 == infoPdf->revision ) {
        //Verifico se la user_password  nulla per poter usare il reverse
        if( !isUserPswd && isUserPasswordV2R3() ) {
            isUserPswd = true;
            userPswd = "";
        }
        SetCharset(chst);
        SetState(min_pw,max_pw,st);
        //Faccio partire il ciclo di cracking
        while ( !isOwnerPasswordV2R3() )
            SetNextPswd();
    }
    string s = GetCurrentPswd();
    clearPassState ();
    return s;
}

void PDFSec :: SetCharset ( const string chs ) {
    if(charSet)
        delete [] charSet;
    lenSet = chs.length();
    charSet = new uint8_t[lenSet];
    for(unsigned i = 0; i < lenSet; i++)
        charSet[i] = (uint8_t)chs[i];
}

void PDFSec :: SetState ( const unsigned min = 1, const unsigned max = 5, unsigned * st = NULL ) {
    //Verifico gli estremi di generazione, se non son validi
    //imposto gli estremi da 1 a 5 di default
    if( min > max && 1 < max && 1 < min ) { min_c = 1; max_c = 5; }
    else { max_c = max; min_c = min; }
    //Genero il vettore degli stati e lo carico (o lo azzero)
    if(state)
        delete [] state;
    state = new unsigned[max_c];
    if( st ) {
        //Copio lo stato
        for(unsigned i = 0; i < max_c; i++)
            state[i] = st[i];
        //Determino la lunghezza
        for(unsigned i = 0; i < max_c; i++)
            if(state[i] == -1)
                lenState = i;
    }
    else {
        for(unsigned i = 0; i < min_c; i++)
            state[i] = 0;
        for(unsigned i = min_c; i < max_c; i++)
            state[i] = -1;
        lenState = min_c;
        
    }
    //Imposto la char
    unsigned i;
    for(unsigned i = 0; i < lenState; i++)
        key[i] = charSet[state[i]];
    for(unsigned i = 0 ; i < 32 - lenState; i++)
        key[lenState + i] = pad[i];
    EOS = false;
}

//Questa funzione lavora direttamente sul vettore key (per risparimiare lavoro)
void PDFSec :: SetNextPswd () {
    //Incremento la combinazione di 1
    register unsigned i = 0;
    while ( i < lenState ) {
        if( ++state[i] < lenSet ) {
            key[i] = charSet[state[i++]];
            break;
        }
        else {
            //Ciclo di password finito
            if( i == max_c ) {
                EOS = true;
                break;
            }
            //Sono all'ultima combinazione con i cifre
            state[i] = 0;
            key[i++] = charSet[0];
            if( i == lenState ) {
                state[lenState] = 0;
                key[lenState] = charSet[0];
                lenState++;
                for(unsigned j = 0; j < 32 - lenState; j++)
                    key[lenState + j] = pad[j];
                break;
            }
        }
    }
}

string PDFSec :: GetCurrentPswd () {
    string str;
    for(unsigned i = 0; i < lenState; i++)
        str.push_back( charSet[state[i]] );
    return str;
}

void PDFSec :: clearCracking() {
    EOS = false;
    isUserPswd = false;
    userPswd = "";
    if(key) {
        delete [] key;
        key = NULL;
    }
    if(u_string) {
        delete [] u_string;
        u_string = NULL;
    }
    if(cifr) {
        delete [] cifr;
        cifr = NULL;
    }
    if(tmp) {
        delete [] tmp;
        tmp = NULL;
    }
    if(tmpkey) {
        delete [] tmpkey;
        tmpkey = NULL;
    }
    if(digest) {
        delete [] digest;
        digest = NULL;
    }
}

void PDFSec :: clearPassState () {
    EOS = false;
    if(state) {
        delete [] state;
        state = NULL;
    }
    if(charSet) {
        delete [] charSet;
        charSet = NULL;
    }
    lenSet = 0;
    max_c = 0;
    min_c = 0;
    lenState = 0;
}
